"""Unicode conformance testing.

- https://www.unicode.org/Public/15.0.0/ucd/auxiliary/GraphemeBreakTest.html
- https://www.unicode.org/Public/15.0.0/ucd/auxiliary/GraphemeBreakTest.txt
- https://www.unicode.org/reports/tr29/tr29-41.html
"""

import pathlib
import time
import unittest

from pyuegc import EGC, UNICODE_VERSION

# Unicode conformance test file
UNICODE_FILE = "GraphemeBreakTest.txt"


def parse_file():
    records = []

#   path = pathlib.Path(__file__).parent / "data" / UNICODE_FILE
    path = pathlib.Path.cwd() / "data" / UNICODE_FILE
    with path.open(encoding="utf-8") as f:
        assert UNICODE_VERSION in f.readline(), "Wrong Unicode version number."
        f.seek(0)

        for num, line in enumerate(f, 1):
            if "#" in line:
                pattern = line.split("#", 1)[0].rstrip()
            if not pattern:
                continue

            cp = []
            breakpoints = []
            cp_count = 0
            for x in pattern.split():
                # ÷ wherever there is a break opportunity
                # × wherever there is not
                if x == "÷":
                    breakpoints.append(cp_count)
                elif x == "×":
                    pass
                else:
                    cp.append(chr(int(x, 16)))
                    cp_count += 1

            string = "".join(cp)
            graphemes = [
                string[i:j]
                for i, j in zip(breakpoints, breakpoints[1:])
            ]
            records.append([num, string, graphemes])

    return records


def make_function(observed, expected):
    def ghost(self):
        self.assertEqual(observed, expected)
    return ghost


class TestExtendedGraphemeClusters(unittest.TestCase):
    pass


def generator():
    for record in parse_file():
        testname = f"test_line_{record[0]:03d}"
        observed = EGC(record[1])
        expected = record[2]
        testfunc = make_function(observed, expected)
        setattr(TestExtendedGraphemeClusters, testname, testfunc)


generator()


if __name__ == "__main__":
    unittest.main()



"""A python implementation of the Unicode algorithm for breaking strings
of text (i.e., code point sequences) into extended grapheme clusters. This
package supports version 15.0 of the Unicode standard (September 13, 2022).
It has been thoroughly tested against the Unicode test file found
at https://www.unicode.org/Public/15.0.0/ucd/auxiliary/GraphemeBreakTest.txt

To get the version of the Unicode character database currently used:

    >>> from pyuegc import UCD_VERSION
    >>> UCD_VERSION
    '15.0.0'

For the formal specification, see Unicode Standard Annex #29, "Unicode Text
Segmentation", at https://www.unicode.org/reports/tr29/tr29-41.html
"""

import sys
if sys.version_info < (3, 6):
    raise SystemExit(f"\n{__package__.title()} requires Python 3.6 or later.")
del sys

__all__ = [
    "EGC",
    "UCD_VERSION",
    "UNICODE_VERSION",
    "__version__",
]

# Unicode standard used to process the data
# Version released on September 13, 2022
UNICODE_VERSION = UCD_VERSION = "15.0.0"


from pyuegc import _version
__version__ = _version.__version__
del _version

from pyuegc.unicode import UNICODE_VERSION as _UNICODE
if _UNICODE != UNICODE_VERSION:
    raise SystemExit(f"\nWrong Unicode version number in {unicode.__name__}")

from pyuegc.egc import EGC



"""Unicode default extended grapheme clusters."""

from pyuegc.unicode import (
    _GCB_PROP_VAL,  # grapheme cluster break property values
    _EXT_PICTOGR,   # extended pictographic characters
)

_GCB_PROP_VAL.update(dict.fromkeys(_EXT_PICTOGR, "Extended_Pictographic"))

# Grapheme cluster break chart
# https://www.unicode.org/Public/15.0.0/ucd/auxiliary/GraphemeBreakTest.html
_GCB_CHART = [
    [1, 1, 1, 1, 0, 1, 1, 0, 1, 1, 1, 1, 1, 1, 0],  # Other
    [1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],  # CR
    [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],  # LF
    [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],  # Control
    [1, 1, 1, 1, 0, 1, 1, 0, 1, 1, 1, 1, 1, 1, 0],  # Extend
    [1, 1, 1, 1, 0, 0, 1, 0, 1, 1, 1, 1, 1, 1, 0],  # Regional_Indicator
    [0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],  # Prepend
    [1, 1, 1, 1, 0, 1, 1, 0, 1, 1, 1, 1, 1, 1, 0],  # SpacingMark
    [1, 1, 1, 1, 0, 1, 1, 0, 0, 0, 1, 0, 0, 1, 0],  # L
    [1, 1, 1, 1, 0, 1, 1, 0, 1, 0, 0, 1, 1, 1, 0],  # V
    [1, 1, 1, 1, 0, 1, 1, 0, 1, 1, 0, 1, 1, 1, 0],  # T
    [1, 1, 1, 1, 0, 1, 1, 0, 1, 0, 0, 1, 1, 1, 0],  # LV
    [1, 1, 1, 1, 0, 1, 1, 0, 1, 1, 0, 1, 1, 1, 0],  # LVT
    [1, 1, 1, 1, 0, 1, 1, 0, 1, 1, 1, 1, 1, 1, 0],  # Extended_Pictographic
    [1, 1, 1, 1, 0, 1, 1, 0, 1, 1, 1, 1, 1, 1, 0],  # ZWJ
]

_GCB_VALUES = (
    # "Other",  # not a GCB property value (used to represent any code point)
    None,       # used in lieu of "Other"
    "CR",
    "LF",
    "Control",
    "Extend",
    "Regional_Indicator",
    "Prepend",
    "SpacingMark",
    "L",
    "V",
    "T",
    "LV",
    "LVT",
    "Extended_Pictographic",  # not a GCB property value
    "ZWJ",
)

_EXT_VALUES = {
    "Extend",
    "Extended_Pictographic",
}


def _nobreak_rules():
    nobreak = []
    for i, sublist in enumerate(_GCB_CHART):
        for j, val in enumerate(sublist):
            if not val:
                nobreak.append((_GCB_VALUES[i], _GCB_VALUES[j]))
    return set(nobreak)


_NOBREAK = _nobreak_rules()


del _EXT_PICTOGR, _GCB_CHART, _GCB_VALUES


def EGC(unistr):
    """Split the original Unicode string *unistr* to a list
    of extended grapheme clusters.
    """
    values = [_GCB_PROP_VAL.get(ord(u)) for u in unistr]

    if values == [None] * len(values):
        return [*unistr] if unistr else []

    # Grapheme breakpoints
    breakpoints = []
    stack = []
    prev = ext_pict_idx = None
    ri_count = 0
    for i, curr in enumerate(values):
        if not i:
            breakpoints.append(1)
        else:
            if prev == "Extended_Pictographic":
                ext_pict_idx = i - 1
            if curr == "Extended_Pictographic" and prev == "ZWJ":
                # https://www.unicode.org/reports/tr29/tr29-41.html#GB11
                # Do not break within emoji modifier sequences
                # or emoji zwj sequences.
                if ext_pict_idx is None:
                    breakpoints.append(1)
                else:
                    tmp = [
                        1 if x in _EXT_VALUES else 0
                        for x in stack[ext_pict_idx : i - 1]
                    ]
                    breakpoints.append(0 if tmp and all(tmp) else 1)
            elif curr == "Regional_Indicator" and prev == "Regional_Indicator":
                # https://www.unicode.org/reports/tr29/tr29-41.html#GB12
                # https://www.unicode.org/reports/tr29/tr29-41.html#GB13
                # Do not break within emoji flag sequences. That is, do not
                # break between regional indicator (RI) symbols if there is
                # an odd number of RI characters before the break point.
                ri_count += 1
                breakpoints.append(0 if ri_count & 1 else 1)
            else:
                breakpoints.append(0 if (prev, curr) in _NOBREAK else 1)

        prev = curr
        stack.append(prev)

    breakpoints.append(1)

    # Grapheme clusters
    egc = []
    prev = 0
    for i, j in enumerate(breakpoints):
        if j and i:
            egc.append(unistr[prev:i])
            prev = i

    return egc



"""Setup script for pyuegc."""

from setuptools import setup, find_packages

URL = "https://github.com/mlodewijck/pyuegc"


def get_version():
    version_file = "pyuegc/_version.py"
    with open(version_file) as f:
        exec(compile(f.read(), version_file, "exec"))
    return locals()["__version__"]

with open("README.md", encoding="utf-8") as f:
    README = f.read()

setup(
    name="pyuegc",
    version=get_version(),
    description=(
        "An implementation of the Unicode algorithm for breaking code point "
        "sequences into extended grapheme clusters as specified in UAX #29"
    ),
    long_description=README,
    long_description_content_type="text/markdown",
    author="Marc Lodewijck",
    author_email="mlodewijck@gmail.com",
    license="MIT",
    url=URL,
    project_urls={
        "Bug Reports": "{}/issues".format(URL),
        "Source": "{}/".format(URL),
    },
    keywords=[
        "Unicode",
        "Unicode grapheme clusters",
        "extended grapheme cluster",
        "EGC",
        "grapheme cluster",
        "graphemes",
        "segmentation",
    ],
    # Trove classifiers
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Software Development",
        "Topic :: Software Development :: Internationalization",
        "Topic :: Text Processing",
        "Topic :: Utilities",
    ],
    python_requires=">=3.6",
    packages=find_packages(),
    include_package_data=True,
    # All data files matched by MANIFEST.in will get included
    # if they are inside a package directory.
    zip_safe=False,
)
